/*
  ccam ver 4.00
  (ver1.00) new data acquision protocol (request & reply)
  (ver1.10:2002.3.14) restrict marker search area to around ping-pong
                      read field size from 'field_size'
  (ver1.11:2002.3.15) some bug fix
  (ver1.12:2002.3.15) bug fix for coordinate restriction to target area
  (ver1.20:2002.3.23) using two color markers for deciding robot ID & dir
  (ver1.21:2002.3.24) some bug fix
  (ver1.22:2002.4.12) read robot shape specification from file
  (ver2.20:2003.3.04) protocol changed
  (ver4.00:2003.4.15) adjusuted version number to wemu4
 */

/*
  Future Plan (2003.4.16)
    - change coordinate system as 'real' scale [cm]
 */

#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <netdb.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/mman.h>

#define Y_LARGE_MARKER_TH 25     /* threshold size for larger Y-marker */
#define B_LARGE_MARKER_TH 25     /* threshold size for larger B-marker */

#define CREATE_RAW_IMAGE       /* create raw & ext. image at 10th frame
				  for checking color extraction         */

//#define VERBOSE_MESSAGE        /* show verbose message */

#define S_TH 3  // threshold of area distinguishing 'mark' from 'noise'

#include "ccam.h"
#include "cmvision.h"
#include "capture.h"

struct sockaddr_in saddr;
struct hostent *hp;

#define mdis(x1,y1,x2,y2) sqrt((double)(x1-x2)*(double)(x1-x2)+(double)(y1-y2)*(double)(y1-y2))

#define RPORT 10030   /* ccam's coordinate request port = 10030 */

int main(int ac, char **av)
{
  capture cap;
  CMVision vision;
  image_pixel *img;
  CMVision::region *reg;
  CMVision::rectangle rect;
  CMVision::line line;
  int c;
  int fd;
  char buf[1024], buf0[1024];
  int fr;
  int t0, t1;
  struct timeval tv;
  double et;
  int bx, by;
  int nry, nrb;
  int nry_found, nrb_found;
  int nrys, nrbs;
  FILE *fp;
  int y1, y2, u0, v0;
  image_pixel p;
  rgb col;
  int i, j, k;
  int min, md;
  int used[NRMAX];
  int XS, YS, X0, Y0, X1, Y1;
  int X;
  int Smark[NRMAX];
  int swap;
  int rfd;
  struct sockaddr_in sraddr;
  int Yexist[NRMAX], ryx[NRMAX], ryy[NRMAX], sry[NRMAX], Ynum[NRMAX];
  int Bexist[NRMAX], rbx[NRMAX], rby[NRMAX], srb[NRMAX], Bnum[NRMAX];

  int Ydir[NRMAX], Bdir[NRMAX];
  int Yx[NRMAX], Yy[NRMAX];
  int Bx[NRMAX], By[NRMAX];

  int ydir0[NRMAX], bdir0[NRMAX];
  int mky1[NRMAX], mky2[NRMAX], idy[NRMAX];
  int mkb1[NRMAX], mkb2[NRMAX], idb[NRMAX];
  char yrname[NRMAX][1024];
  char brname[NRMAX][1024];
  int Myx[NMKMAX], Myy[NMKMAX];
  int myx[NMKMAX], myy[NMKMAX];
  int sMy[NMKMAX], smy[NMKMAX];
  int Mbx[NMKMAX], Mby[NMKMAX];
  int mbx[NMKMAX], mby[NMKMAX];
  int sMb[NMKMAX], smb[NMKMAX];
  unsigned char yrr[NRMAX]; /* marker search distance of Y */
  unsigned char brr[NRMAX]; /* marker search distance of B */
  int nmy, nMy;
  int nmb, nMb;
  int buflen;
  int Nyrobots, Nbrobots;
  int mk1, mk2, dir, rr, id;
  char team;
  char rname[256];

  if (ac < 4){
    fprintf(stderr, "usage: %s NRY NRB LUT_FILE_NAME\n", av[0]); exit(1);
  }
  nrys = atoi(av[1]);
  nrbs = atoi(av[2]);

  if ((fp = fopen(ROBOT_SHAPE_FILENAME, "r")) == NULL){
    printf("can't open robot shape file '%s'.\n",
	   ROBOT_SHAPE_FILENAME); exit(1);
  }
  Nyrobots = 0;
  Nbrobots = 0;
  while(fgets(buf, 1023, fp) != NULL){
    if (buf[0] != '#'){
      mk1 = atoi(strtok(buf, " ,\t\n"));
      mk2 = atoi(strtok(NULL, " ,\t\n"));
      dir = atoi(strtok(NULL, " ,\t\n"));
      rr = atoi(strtok(NULL, " ,\t\n"));
      team = *strtok(NULL, " ,\t\n");
      id = atoi(strtok(NULL, " ,\t\n"));
      strcpy(rname, strtok(NULL, " ,\t\n"));

      if (team == 'Y' || team == 'y'){
	mky1[Nyrobots] = mk1;
	mky2[Nyrobots] = mk2;
	ydir0[Nyrobots] = dir;
	yrr[Nyrobots] = rr;
	idy[Nyrobots] = id;
	strcpy(yrname[Nyrobots], rname);
	Nyrobots++;
      }
      else{
	mkb1[Nbrobots] = mk1;
	mkb2[Nbrobots] = mk2;
	bdir0[Nbrobots] = dir;
	brr[Nbrobots] = rr;
	idb[Nbrobots] = id;
	strcpy(brname[Nbrobots], rname);
	Nbrobots++;
      }
    }
  }

  if (!cap.initialize("/dev/video",X, Y, V4L2_PIX_FMT_BGR24)){
      fprintf(stderr, "Capture init failed.\n"); exit(1);
  }

  if (!vision.initialize(X, Y)){
      fprintf(stderr, "Vision init failed.\n"); exit(1);
  }

  vision.loadOptions("colors.txt");
  vision.enable(CMV_DENSITY_MERGE);

  if ((fp = fopen(FIELD_SIZE_FILENAME, "r")) == NULL){
    fprintf(stderr, "Error: can't open %s\n", FIELD_SIZE_FILENAME);
    exit(1);
  }
  fgets(buf, 1023, fp); fclose(fp);
  X0 = atoi(strtok(buf, " \t\n"));
  Y0 = atoi(strtok(NULL, " \t\n"));
  X1 = atoi(strtok(NULL, " \t\n"));
  Y1 = atoi(strtok(NULL, " \t\n"));
  if (X0 > X1){ swap = X0; X0 = X1; X1 = swap;}
  if (Y0 > Y1){ swap = Y0; Y0 = Y1; Y1 = swap;}
  XS = X1 - X0;
  YS = Y1 - Y0;

  /* field: (X0, Y0) - (X1, Y1) (size: XS * YS) */

  /* making sockets */
  if ((rfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
      fprintf(stderr, "Error: Failed to make socket for receiving.\n");
      exit(1);
  }

  /* socket for receiving */
  bzero((char *)&sraddr, sizeof(sraddr));
  sraddr.sin_family = AF_INET;
  sraddr.sin_addr.s_addr = INADDR_ANY;
  sraddr.sin_port = htons(RPORT);

  if (bind(rfd, (struct sockaddr *)&sraddr, sizeof(sraddr)) < 0){
      fprintf(stderr, "Error: Failed to bind socket for receiving.\n");
      exit(1);
  }

  vision.set_yuvbits_using_obj_lut(av[3]);

  fr = 0;
  while(1){
    fr++;
    img = (image_pixel*)cap.captureFrame();

#ifdef CREATE_RAW_IMAGE
    if (fr == 10){
      fp = fopen("raw.ppm","wb");
      fprintf(fp,"P6\n%d %d\n255\n", X, Y);
      for (i =0; i < X*Y; i++){
	p = img[i];
	col.red = p.red; col.green = p.green; col.blue = p.blue;
	fprintf(fp,"%c%c%c",col.red,col.green,col.blue);
      }
      fclose(fp);
      fp = fopen("ext.ppm","wb");
      fprintf(fp,"P6\n%d %d\n255\n", X, Y);
      for (i =0; i < X*Y; i++){
	int f0;
	unsigned int r0, g0, b0;
	p = img[i];
	r0 = p.red; g0 = p.green; b0 = p.blue;
	f0 = fl[r0/2][g0/2][b0/2];
	col.red = 0; col.green = 0; col.blue = 0;
	if ((f0 & 0x01)!=0){col.red=0xff; col.green=0x00; col.blue=0x00;}
	if ((f0 & 0x02)!=0){col.red=0xff; col.green=0xff; col.blue=0x00; }
	if ((f0 & 0x04)!=0){col.red=0x00; col.green=0x00; col.blue=0xff; }
	if ((f0 & 0x08)!=0){col.red=0x00; col.green=0xff; col.blue=0x00; }
	if ((f0 & 0x10)!=0){col.red=0xff; col.green=0x00; col.blue=0xff; }
	if ((f0 & 0x20)!=0){col.red=0x00; col.green=0xff; col.blue=0xff; }
	if ((f0 & 0x40)!=0){col.red=0xff; col.green=0x80; col.blue=0x80; }
	fprintf(fp,"%c%c%c",col.red,col.green,col.blue);
      }
      fclose(fp);
    }
#endif
    vision.processFrame(img);

    gettimeofday(&tv, NULL);
    t1 = tv.tv_usec;
    et = (double)(t1-t0)/1000.0; if (et < 0.0) et += 1000.0;
    t0 = t1;

    int ba;
    int flag;
    int n;
    int Rx, Ry;
    int M1, M2;
    int M1x, M1y, M2x, M2y;
    int rn;

    /*
      Region code:
       0 = ball
       1 = Y-Ping
       2 = B-Ping
       3 = Y-Marker1
       4 = Y-Marker2
       5 = B-Marker1
       6 = B-Marker2
     */

    /* search for Ball */
    flag = 0;
    reg = vision.getRegions(0);
    if (vision.numRegions(0) > 0){
      while(flag == 0){
	bx = (int)reg->cen_x-X0; by = YS - (int)reg->cen_y + Y0;
	ba = reg->area;
	if (bx >= 0 && bx <= XS && by >= 0 && by <= YS && ba >= S_TH) flag = 1;
	else if (reg->next == NULL){
	  bx = -1; by = -1; flag = 1;
	}
	else reg = reg->next;
      }
    }
    else{
      bx = -1; by = -1; // note: should be ball's previous position ?
    }
#ifdef VERBOSE_MESSAGE
    printf("Ball:(%d %d : %d)\n", bx, by, ba);
#endif
    /* search for Y-Ping */
    nry = 0;
    if (vision.numRegions(1) > 0){
      reg = vision.getRegions(1);
      while(nry < nrys){
	flag = 0;
	while(flag == 0 && reg != NULL){
	  ryx[nry] = (int)reg->cen_x-X0; ryy[nry] = YS - (int)reg->cen_y + Y0;
	  /* found inside field ? */
	  if (ryx[nry] >= 0 && ryx[nry] <= XS
	      && ryy[nry] >= 0 && ryy[nry] <= YS) flag = 1;
	  reg = reg->next;
	  if (reg == NULL){
	    if (flag == 0){
	      ryx[nry] = -1;
	      ryy[nry] = -1; // note: should be previous position ?
	      flag = 1;
	    }
	  }
	}
	nry++;
      }
    }

    for (i = 1; i <= NRMAX; i++) Yexist[i] = 0;

    if (Nyrobots != 0){ /* in case of using Y-markers ... */
      /* search for Y-marker1&2 */
      n = vision.numRegions(3); /* Y-marker1 */
      nMy = 0;
      reg = vision.getRegions(3);
      for (i = 0; i < n; i++){
	Myx[nMy] = (int)reg->cen_x-X0; Myy[nMy] = YS - (int)reg->cen_y + Y0;
	sMy[nMy] = (int)reg->area;
	if (Myx[nMy] >= 0 && Myx[nMy] <= XS
	    && Myy[nMy] >= 0 && Myy[nMy] <= YS && sMy[nMy] >= S_TH){
	  nMy++;
	}
	reg = reg->next;
      }
      
      n = vision.numRegions(4); /* Y-marker2 */
      nmy = 0;
      reg = vision.getRegions(4);
      for (i = 0; i < n; i++){
	myx[nmy] = (int)reg->cen_x-X0; myy[nmy] = YS-(int)reg->cen_y+Y0;
	smy[nmy] = (int)reg->area;
	if (myx[nmy] >= 0 && myx[nmy] <= XS
	    && myy[nmy] >= 0 && myy[nmy] <= YS && smy[nmy] >= S_TH){
	  nmy++;
	}
	reg = reg->next;
      }
      
      nry_found = 0;
      for (i = 0; i < nry; i++){
#ifdef VERBOSE_MESSAGE
	printf("  Y: for #%d (%d %d): ", i, ryx[i], ryy[i]);
#endif
	Rx = ryx[i]; Ry = ryy[i];
	M1 = 0; M2 = 0;
	for (j = 0; j < nMy; j++){
	  if (mdis(Myx[j], Myy[j], Rx, Ry) < yrr[i]){ /* Y-Marker1 exists? */
	    if (sMy[j] > Y_LARGE_MARKER_TH){ /* is it large Y-Marker1? */
	      M1 = 2; M1x = Myx[j]; M1y = Myy[j];
#ifdef VERBOSE_MESSAGE
	      //	      printf("M1(%d:%.1f) ", sMy[j], mdis(Myx[j],Myy[j],Rx,Ry));
	      printf("M1(%d:%.1f)@(%d %d)",
		     sMy[j], mdis(Myx[j],Myy[j],Rx,Ry),
		     Myx[j], Myy[j]);
#endif
	  }
	    else{ /* it is small Y-Marker1 */
	      M1 = 1; M1x = Myx[j]; M1y = Myy[j];
#ifdef VERBOSE_MESSAGE
	      printf("m1(%d:%.1f) ", sMy[j], mdis(Myx[j],Myy[j],Rx,Ry));
#endif
	    }
	  }
	}
	for (j = 0; j < nmy; j++){
	  if (mdis(myx[j], myy[j], Rx, Ry) < yrr[i]){ /* Y-Marker2 exists? */
	    if (smy[j] > Y_LARGE_MARKER_TH){ /* is it large Y-Marker2? */
	      M2 = 2; M2x = myx[j]; M2y = myy[j];
#ifdef VERBOSE_MESSAGE
	      printf("M2(%d:%.1f) ", smy[j], mdis(myx[j],myy[j],Rx,Ry));
#endif
	    }
	    else{ /* it is small Y-Marker2 */
	      M2 = 1; M2x = myx[j]; M2y = myy[j];
#ifdef VERBOSE_MESSAGE
	      printf("m2(%d:%.1f) ", smy[j], mdis(myx[j],myy[j],Rx,Ry));
#endif
	    }
	  }
	}
#ifdef VERBOSE_MESSAGE
	printf("\n");
#endif
	
	/* search for Y robots from shape.txt description */
	for (j = 0; j < Nyrobots; j++){
	  if (mky1[j] == M1 && mky2[j] == M2){
	    rn = idy[j];
	    Yexist[rn] = 1; Yx[rn] = Rx; Yy[rn] = Ry; Ynum[rn] = j;
	    nry_found++;
	    if (mky1[j] != 0 && mky2[j] != 0){
	      Ydir[rn] = (int)(-atan2((double)(M1y-M2y),(double)(M1x-M2x))
		/ M_PI * 180.0) + ydir0[j];
	    }
	    else if (mky1[j] != 0){
	      Ydir[rn] = (int)(-atan2((double)(M1y-Ry),(double)(M1x-Rx))
		/ M_PI * 180.0) + ydir0[j];
	    }
	    else{
	      Ydir[rn] = (int)(-atan2((double)(Ry-M2y),(double)(Rx-M2x))
		/ M_PI * 180.0)  + ydir0[j];
	    }
	    /* truncate direction into [0:360] */
	    while(Ydir[rn] >= 360) Ydir[rn] -= 360;
	    while(Ydir[rn] < 0) Ydir[rn] += 360;
	    break;
	  }
	}
      }
    }
    else{ /* in case of not using Y-markers... */
      for (i = 0; i < nry; i++){
	Yexist[i+1] = 1;
	Ydir[i+1] = -1;
      }
    }
    /* search for B-Ping */
    nrb = 0;
    if (vision.numRegions(2) > 0){
      reg = vision.getRegions(2);
      while(nrb < nrbs){
	flag = 0;
	while(flag == 0 && reg != NULL){
	  rbx[nrb] = (int)reg->cen_x-X0; rby[nrb] = YS-(int)reg->cen_y+Y0;
	  srb[nrb] = (int)reg->area;
	  /* found inside field ? */
	  if (rbx[nrb] >= 0 && rbx[nrb] <= XS
	      && rby[nrb] >= 0 && rby[nrb] <= YS && srb[nrb] > S_TH) flag = 1;
	  reg = reg->next;
	  if (reg == NULL){
	    if (flag == 0){
	      rbx[nrb] = -1;
	      rby[nrb] = -1; // note: should be previous position ?
	      flag = 1;
	    }
	  }
	}
	nrb++;
      }
    }

    for (i = 1; i <= NRMAX; i++) Bexist[i] = 0;

    if (Nbrobots != 0){ /* in case of using B-markers... */
      /* search for B-marker1&2 */
      n = vision.numRegions(5); /* B-marker1 */
      nMb = 0;
      reg = vision.getRegions(5);
      for (i = 0; i < n; i++){
	Mbx[nMb] = (int)reg->cen_x-X0; Mby[nMb] = YS-(int)reg->cen_y+Y0;
	sMb[nMb] = (int)reg->area;
	if (Mbx[nMb] >= 0 && Mbx[nMb] <= XS
	    && Mby[nMb] >= 0 && Mby[nMb] <= YS && sMb[nMb] >= S_TH){
	  nMb++;
	}
	reg = reg->next;
      }

      n = vision.numRegions(6); /* B-marker2 */
      nmb = 0;
      reg = vision.getRegions(6);
      for (i = 0; i < n; i++){
	mbx[nmb] = (int)reg->cen_x-X0; mby[nmb] = YS-(int)reg->cen_y+Y0;
	smb[nmb] = (int)reg->area;
	if (mbx[nmb] >= 0 && mbx[nmb] <= XS
	    && mby[nmb] >= 0 && mby[nmb] <= YS && smb[nmb] >= S_TH){
	  nmb++;
	}
	reg = reg->next;
      }

      nrb_found = 0;
      for (i = 0; i < nrb; i++){
#ifdef VERBOSE_MESSAGE
	printf("  B: for #%d (%d %d): ", i, rbx[i], rby[i]);
#endif
	Rx = rbx[i]; Ry = rby[i];
	M1 = 0; M2 = 0;
	for (j = 0; j < nMb; j++){
	  if (mdis(Mbx[j], Mby[j], Rx, Ry) < brr[i]){ /* B-Marker1 exists? */
	    if (sMb[j] > B_LARGE_MARKER_TH){ /* is it large B-Marker1? */
	      M1 = 2; M1x = Mbx[j]; M1y = Mby[j];
#ifdef VERBOSE_MESSAGE
	      printf("M1(%d:%.1f) ", sMb[j], mdis(Mbx[j],Mby[j],Rx,Ry));
#endif
	    }
	    else{ /* it is small B-Marker1 */
	      M1 = 1; M1x = Mbx[j]; M1y = Mby[j];
#ifdef VERBOSE_MESSAGE
	      printf("m1(%d:%.1f) ", sMb[j], mdis(Mbx[j],Mby[j],Rx,Ry));
#endif
	    }
	  }
	}
	for (j = 0; j < nmb; j++){
	  if (mdis(mbx[j], mby[j], Rx, Ry) < brr[i]){ /* B-Marker2 exists? */
	    if (smb[j] > B_LARGE_MARKER_TH){ /* is it large B-Marker2? */
	      M2 = 2; M2x = mbx[j]; M2y = mby[j];
#ifdef VERBOSE_MESSAGE
	      printf("M2(%d:%.1f) ", smb[j], mdis(mbx[j],mby[j],Rx,Ry));
#endif
	    }
	    else{ /* it is small B-Marker2 */
	      M2 = 1; M2x = mbx[j]; M2y = mby[j];
#ifdef VERBOSE_MESSAGE
	      printf("m2(%d:%.1f) ", smb[j], mdis(mbx[j],mby[j],Rx,Ry));
#endif
	    }
	  }
	}
#ifdef VERBOSE_MESSAGE
	printf("\n");
#endif
	
	/* search for B robots from shape.txt description */
	for (j = 0; j < Nbrobots; j++){
	  if (mkb1[j] == M1 && mkb2[j] == M2){
	    nrb_found++;
	    rn = idb[j];
	    Bexist[rn] = 1; Bx[rn] = Rx; By[rn] = Ry; Bnum[rn] = j;
	    if (mkb1[j] != 0 && mkb2[j] != 0){
	      Bdir[rn] = int(-atan2((double)(M1y-M2y),(double)(M1x-M2x))
		/ M_PI * 180.0) + bdir0[j];
	    }
	    else if (mkb1[j] != 0){
	      Bdir[rn] = int(-atan2((double)(M1y-Ry),(double)(M1x-Rx))
		/ M_PI * 180.0) + bdir0[j];
	    }
	    else{
	      Bdir[rn] = int(-atan2((double)(Ry-M2y),(double)(Rx-M2x))
		/ M_PI * 180.0)  + bdir0[j];
	    }
	    /* truncate direction into [0:360] */
	    while(Bdir[rn] >= 360) Bdir[rn] -= 360;
	    while(Bdir[rn] < 0) Bdir[rn] += 360;
	    break;
	  }
	}
      }
    }
    else{ /* in case of not using B-markers... */
      for (i = 0; i < nrb; i++){
	Bx[i+1] = rbx[i]; By[i+1] = rby[i];
	Bexist[i+1] = 1;
	Bdir[i+1] = -1;
      }
    }

#ifdef VERBOSE_MESSAGE
    printf("Yrobots(\n");
    for (i = 1; i <= NRMAX; i++)
      if (Yexist[i] != 0){
	printf("  ID=%d(%4s) at (%d %d / %d)\n",
	       i, yrname[Ynum[i]], Yx[i], Yy[i], Ydir[i]);
      }
    printf(")\n");

    printf("Brobots:(\n");
    for (i = 1; i <= NRMAX; i++)
      if (Bexist[i] != 0){
	printf("  ID=%d(%s) at (%d %d /%d)\n",
	       i, brname[Bnum[i]], Bx[i], By[i], Bdir[i]);
      }
    printf(")\n");
#endif

    nry_found = 0; nrb_found = 0;
    for(i=1;i<=NRMAX;i++){
      if (Yexist[i]!=0)nry_found++;
      if (Bexist[i]!=0)nrb_found++;
    }

    printf("Sb=%d Y=%d B=%d #%d [%d %d %d %d %d %d %d] (%.1f)\n",
	   ba, nry_found, nrb_found, fr,
	   vision.numRegions(0),
	   vision.numRegions(1), vision.numRegions(2),
	   vision.numRegions(3), vision.numRegions(4), 
	   vision.numRegions(5), vision.numRegions(6),
	   1000.0/et);

    fd_set rfds;
    struct timeval tv;
    int len;
    int ret;
    struct sockaddr_in caddr;
    char shost[1024];
    int sport;
    int Tid;

    tv.tv_sec = 0; tv.tv_usec = 0;
    FD_ZERO(&rfds);
    FD_SET(rfd, &rfds);
    if (select(64, &rfds, (fd_set *)0, (fd_set *)0, &tv) < 0) {
	fprintf(stderr, "Warning: failed to select(2) in receive().\n");
	exit(1);
    }

    if(FD_ISSET(rfd, &rfds)){
      len = sizeof(caddr);
      if ((ret = recvfrom(rfd, buf, 1024, 0, 
			  (struct sockaddr *)&caddr,(socklen_t *)&len))<0){
	fprintf(stderr,
		"Warning: failed to recvfrom(2) in receive().\n");
	exit(1);
      }

      /*
	coordinate request form:
	  hostname port team robot-ID
      */
      buf[ret] = '\0';
      strcpy(shost, strtok(buf, " \t\n"));
      sport = atoi(strtok(NULL, " \t\n"));
      team = *strtok(NULL, " \t\n");
      Tid = atoi(strtok(NULL, " \t\n"));
      if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	perror("socket"); exit(1);
      }
      bzero((char *)&saddr, sizeof(saddr));
      if ((hp = gethostbyname(shost)) == NULL) {
	perror("No such host"); exit(1);
      }
      bcopy(hp->h_addr, &saddr.sin_addr, hp->h_length);
      saddr.sin_family = AF_INET;
      saddr.sin_port = htons(sport);

      sprintf(buf, "%d %d %d %d %d ", fr, XS, YS, bx, by);

      if (team == 'Y' || team == 'y'){ /* when request for Y team */
	sprintf(buf0, "%d %d ", nry_found, nrb_found); strcat(buf, buf0);
	if (Yexist[Tid] != 0){
	  sprintf(buf0, "%d %d %d ", Yx[Tid], Yy[Tid], Ydir[Tid]);
	}
	else{
	  sprintf(buf0, "-1 -1 -1 ");
	}
	strcat(buf, buf0);
	for (i = 1; i <= NRMAX; i++){
	  if (Yexist[i] != 0 && i != Tid){
	    sprintf(buf0, "%d %d %d ", Yx[i], Yy[i], Ydir[i]);
	    strcat(buf, buf0);
	  }
	}
	for (i = 1; i <= NRMAX; i++){
	  if (Bexist[i] != 0){
	    sprintf(buf0, "%d %d %d ", Bx[i], By[i], Bdir[i]);
	    strcat(buf, buf0);
	  }
	}
      }
      else{ /* when request for B team */
	sprintf(buf0, "%d %d ", nrb_found, nry_found); strcat(buf, buf0);
	if (Bexist[Tid] != 0){
	  sprintf(buf0, "%d %d %d ", Bx[Tid], By[Tid], Bdir[Tid]);
	}
	else{
	  sprintf(buf0, "-1 -1 -1 ");
	}
	strcat(buf, buf0);
	for (i = 0; i <= NRMAX; i++){
	  if (Bexist[i] != 0 && i != Tid){
	    sprintf(buf0, "%d %d %d ", Bx[i], By[i], Bdir[i]);
	    strcat(buf, buf0);
	  }
	}
	for (i = 0; i <= NRMAX; i++){
	  if (Yexist[i] != 0){
	    sprintf(buf0, "%d %d %d ", Yx[i], Yy[i], Ydir[i]);
	    strcat(buf, buf0);
	  }
	}
      }
      strcat(buf, "\n");
      
      buflen = strlen(buf);
      if (sendto(fd, buf, buflen, 0,
		 (struct sockaddr *)&saddr, sizeof(saddr)) != buflen){
	perror("sendto"); exit(1);
      }
      close(fd);
    }
  }
  cap.close();
  vision.close();
  close(fd);
}
