/*
  ssend ver4.00
    (c)Junichi Akita (akita@fun.ac.jp)
*/

#include "ccam.h"

#define MAX_SPEED (255)

struct termios tio;

#define DEBUG

#define iabs(x) ((x>0)?x:(-x))

unsigned char *set_data(unsigned char *c, int id, int L, int R)
{
  int i;
  *(c+(id*3)+1) = iabs(R) & 0xff;
  *(c+(id*3)+2) = iabs(L) & 0xff;
  if (L < 0) *(c+(id*3)+3) &= ~0x10; else *(c+(id*3)+3) |= 0x10;
  if (R < 0) *(c+(id*3)+3) &= ~0x20; else *(c+(id*3)+3) |= 0x20;
  for (i = 0; i < 5; i++) *(c+(i*3)+3) &= ~0x01;
  *(c+(id*3)+3) |= 0x01;
  return(c);
}


main(int ac, char **av)
{
  unsigned char *c;
  int cfd, sfd;
  struct timeval tv;
  char line[1024];
  fd_set rfds;
  int ret, len;
  struct sockaddr_in sraddr;
  struct sockaddr_in caddr;
  double lm0, rm0, cm0;
  int lm, rm, cm;
  int rn;
  int dlen;
  int i;
  int ty;
  double sv10, sv20, sv30, m10, m20, m30;
  int sv1, sv2, sv3, m1, m2, m3;
  unsigned char mc1, mc2, mc3;
  char rname[NRMAX][256];
  int id0;
  char rname0[256];
  char team;
  int t0;
  FILE *fp;
  int Nrobots;
  char buf[1024];

#define TEAM 'Y'
// #define TEAM 'B'


  if (ac != 2){
    fprintf(stderr, "usage: %s RPORT\n", av[0]); exit(1);
  }
  if ((sfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    fprintf(stderr, "Error: Failed to make socket for receiving.\n");
    exit(1);
  }

  if ((fp = fopen(ROBOT_SHAPE_FILENAME, "r")) == NULL){
    printf("can't open robot shape file '%s'.\n",
	   ROBOT_SHAPE_FILENAME); exit(1);
  }
  for (i = 1; i <= NRMAX; i++) strcpy(rname[i], "---");
  while(fgets(buf, 1023, fp) != NULL){
    if (buf[0] != '#'){
      strtok(buf, " ,\t\n");
      strtok(NULL, " ,\t\n");
      strtok(NULL, " ,\t\n");
      strtok(NULL, " ,\t\n");
      team = *strtok(NULL, " ,\t\n");
      id0 = atoi(strtok(NULL, " ,\t\n"));
      strcpy(rname0, strtok(NULL, " ,\t\n"));
      if (team == TEAM){
	strcpy(rname[id0], rname0);
      }
    }
  }


  bzero((char *)&sraddr, sizeof(sraddr));
  sraddr.sin_family = AF_INET;
  sraddr.sin_addr.s_addr = INADDR_ANY;
  sraddr.sin_port = htons(atoi(av[1]));

  if (bind(sfd, (struct sockaddr *)&sraddr, sizeof(sraddr)) < 0){
    fprintf(stderr, "Error: Failed to bind socket for receiving.\n");
    exit(1);
  }

  cfd = open("/dev/ttyS0", O_RDWR | O_NOCTTY);
  if (cfd < 0){perror("/dev/ttyS0"); exit(-1);}

  bzero((char *)&tio, sizeof(tio));

  tio.c_cflag = B9600 | CRTSCTS | CS8 | CLOCAL | CREAD;
  tio.c_iflag = IGNPAR | ICRNL;
  tio.c_iflag = IGNPAR;
  tio.c_oflag = 0;
  tio.c_lflag = ICANON;
  fcntl(cfd, F_SETFL, FNDELAY);
  tcflush(cfd, TCIFLUSH);
  tcsetattr(cfd, TCSANOW, &tio);

  c = (unsigned char *)malloc(sizeof(unsigned char) * 16);
  c[0] = 'B';
  for (i = 0; i < 5; i++){
    *(c+(i*3)+1) = 0x00;
    *(c+(i*3)+2) = 0x00;
    *(c+(i*3)+3) = 0x30;
  } 

  rn = 0;
  while(1){
    tv.tv_sec = 0; tv.tv_usec = 0;
    FD_ZERO(&rfds);
    FD_SET(sfd, &rfds);
    if (select(64, &rfds, (fd_set *)0, (fd_set *)0, &tv) < 0) {
      fprintf(stderr, "Warning: failed to select(2) in receiveUDPPacket().\n");
      exit(1);
    }
    if(FD_ISSET(sfd, &rfds)){
      len = sizeof(caddr);
      if ((ret = recvfrom(sfd, line, 1024, 0, 
			  (struct sockaddr *)&caddr, (socklen_t *)&len)) < 0){
	fprintf(stderr,
		"Warning: failed to recvfrom(2) in receiveUDPPacket().\n");
	exit(1);
      }
      line[ret] = '\0';
      if (strlen(line) > 0){
#ifdef DEBUG
	printf("*received string: %s / ",line);
#endif
	rn = atoi(strtok(line, " \t\n"));

	lm0 = atof(strtok(NULL, " \t\n"));
	rm0 = atof(strtok(NULL, " \t\n"));

	if (lm0 > 1.0){
	  printf("warning: too large number: L=%f\n", lm0); lm0 = 1.0;
	}
	if (lm0 < -1.0){
	  printf("warning: too large number: L=%f\n", lm0); lm0 = -1.0;
	}
	if (rm0 > 1.0){
	  printf("warning: too large number: R=%f\n", rm0); rm0 = 1.0;
	}
	if (rm0 < -1.0){
	  printf("warning: too large number: R=%f\n", rm0); rm0 = -1.0;
	}
	lm = (int)(lm0 * (double)MAX_SPEED);
	rm = (int)(rm0 * (double)MAX_SPEED);

	c = set_data(c, rn-1, lm, rm);
	printf("#%d:%s:[%4.1f %4.1f]->[%2d %2d]\n",
	       rn, rname[rn], lm0, rm0, lm, rm);
	for (i = 0; i < 16; i++){
	  write(cfd, c+i, 1);
#ifdef DEBUG
	  printf("%02x ", *(c+i));
#endif
	}
#ifdef DEBUG
	printf("\n");
#endif
      }
    }
  }
}
